#include "fscasio.h"
#include <ctype.h>
#include <stdlib.h>
#include <string.h>


#define PAGESTATUS_NORMAL 		0x0000
#define PAGESTATUS_UNUSED 		0xFFFF
#define PAGESTATUS_FILETABLE	0x00FF

#define HEADER_SIZE				0x20
#define PREHEADER_SIZE			0x08
#define FILENAME_SIZE           0x18

#define ROOT_ID					0xFFFF

//First file header byte value :
#define FILE_HEADER_DELETED		0x01
#define FILE_HEADER_START		0x51
//Second file header byte value :
#define FILE_HEADER_FILE		0x20
#define FILE_HEADER_POSTFILE	0x30
#define FILE_HEADER_DIR			0x10



#define CHAR_CASE_NOCHANGE		0
#define CHAR_CASE_UPPER			1
#define CHAR_CASE_AUTO			2


const struct file_operations fscasio_fop = {
	.remove=fscasio_remove,
	.rename=fscasio_rename,
	.fopen=fscasio_fopen,
	.fclose=fscasio_fclose,
	.fgetc=fscasio_fgetc,
	.fgets=fscasio_fgets,
	.fputc=fscasio_fputc,
	.fputs=fscasio_fputs,
	.ungetc=fscasio_ungetc,
	.fread=fscasio_fread,
	.fwrite=fscasio_fwrite,
	.fseek=fscasio_fseek,
	.ftell=fscasio_ftell,
	.fflush=fscasio_fflush
};


//Internals functions prototypes :
/**
  * Return the fisrt file header byte address (at the end of pages header area)
  */
const unsigned char *getFSHeaderAddress();

/**
  * Return the first header byte of the file (or directory) described by the name filename and the parent 
  *  directory ID given by parentDirId (ROOT_ID if the file is in the system root), if existing. Else 
  *  return NULL.
  * WARNING : the file name MUST be "atomic" (isn't a path, it's just the relative file name!)
  */
const unsigned char *getAtomicFileHeader(const char *filename, unsigned short parentDirId);

/**
  * Return the first header byte of the file (or directory) with its path. This path is an absolute path 
  *  (starting with \\fls0\ or /fls0/) that can comport multiple sub-directory, for example "\\fls0\foo\bar.txt".
  * If don't exist, return NULL.
  */
const unsigned char *getFileHeader(const char *filename);

/**
  * Compare the string str to the file/dir header name pointed by fileHeaderName. Return 0 if it match.
  * If str is too long to represent a file name (max : 8 char for the name, 1 char for the dot and 3 char for
  *  the extension), it will be resized using the Casio file name rules, with the suffix "~1".
  * caseRule indicate how to proceding with the character case. It's value may be CHAR_CASE_UPPER, CHAR_CASE_AUTO
  *  or CHAR_CASE_NOCHANGE.
  */
int cmpstr2fhn(const char *str, const unsigned char *fileHeaderName, int caseRule);

/**
  * Return 1 if the fileHeader is a directory header, 0 else.
  */
int isDirectory(const unsigned char *fileHeader);

/**
  * Return 1 if the fileHeader is a deleted file header, 0 else.
  */
int isDeleted(const unsigned char *fileHeader);

/**
  * Return the file (or directory) internal ID.
  */
unsigned short getFileId(const unsigned char *fileHeader);

/**
  * Return the file (or directory) parent directory ID.
  */
unsigned short getFileParentDirId(const unsigned char *fileHeader);

/**
  * Return the number of fragment of file. If it's a directory, return -1.
  */
unsigned short getFileFragNumber(const unsigned char *fileHeader);

/**
  * Return the total size of the file (sum of the size of all the fragments).
  */
unsigned int getFileTotalSize(const unsigned char *fileHeader);

/**
  * Return the offset of a mempage from its ID, or -1 if this ID don't exist yet.
  */
unsigned int getPageOffsetByID(unsigned short id);

/**
  * Return the address of the first data byte of the fragment at address fragAdr. If its mempage
  *  ID don't exist, this function return NULL.
  */
const unsigned char *getFragDataAddress(const unsigned char *fragAdr);


// Functions definitions :

FILE *fscasio_fopen(const char * filename, const char * mode) {
	int i;
	int fval;
	int fileHandle;
	FILE *ret;

	ret = NULL;

	// Check if there is a file location empty
	for (i=0; (i<FOPEN_MAX) && ((fval=(int)__files[i])!=0); i++);
	if(fval == 0) {
		const unsigned char *header;
		fileHandle = i;

		// Parse the open mode
		int m=0;
		if(mode[0] != 0) {
			if(mode[0] == 'r') m |= _FILE_READ;
			else if(mode[0] == 'w') m |= _FILE_WRITE;
			else if(mode[0] == 'a') m |= _FILE_WRITE | _FILE_APPEND;
			if(mode[1] != 0) {
				if(mode[1] == 'b') m |= _FILE_BINARY;
				else if(mode[1] == '+') m |= _FILE_UPDATE | _FILE_WRITE | _FILE_READ;
				if(mode[2] != 0) {
					if(mode[2] == 'b') m |= _FILE_BINARY;
					else if(mode[2] == '+') m |= _FILE_UPDATE | _FILE_WRITE | _FILE_READ;
				}
			}
		}

		//For now, the only mode allowed are "r" or "rb".
		if((m == _FILE_READ) || (m == _FILE_READ | _FILE_BINARY)) {
			header = getFileHeader(filename);
			if((header != NULL) && !isDirectory(header)) {

				FILE *f = malloc(sizeof(FILE));
				__files[i] = f;
				ret = f;
				f->fs_fop = &fscasio_fop;

				f->pos = 0;
				f->open_mode = m;
				f->error = 0;
				f->flags = _FILE_BUF_AUTO;

				f->bufsize = BUFSIZ; // Use the default buffer size
				f->buf = malloc(sizeof(char) * BUFSIZ);
				f->cur_bufsize = 0;  // The buffer is initialy empty
				f->iobf_t = _IOFBF; // Full buffered because isn't an interactive file

				// Private data :
				struct _fscasio_file *pdata = malloc(sizeof(struct _fscasio_file));
				f->private_data = pdata;
				pdata->current_frag = 0;
				pdata->curfrag_pos = 0;
				pdata->index = header;
				pdata->frag_num = getFileFragNumber(header);
				pdata->total_size = getFileTotalSize(header);				
				pdata->frags = malloc(pdata->frag_num * sizeof(struct _fscasio_fragdata));
				int pos = PREHEADER_SIZE + FILENAME_SIZE;
				for(i=0; i<pdata->frag_num; i++, pos+=HEADER_SIZE) {
					pdata->frags[i].fstart = getFragDataAddress((const unsigned char*)((int)(header)+pos));
					pdata->frags[i].fsize = *((unsigned short*)((int)(header)+pos+18))+1;
				}

				if(m & _FILE_APPEND) {
					f->pos = pdata->total_size;
					pdata->current_frag = pdata->frag_num - 1;
					pdata->curfrag_pos = pdata->frags[pdata->current_frag].fsize;
				}
			}
		}
	}
	return ret;
}


int fscasio_fclose(struct _FILE *stream) {
	int i;

	fscasio_fflush(stream);
	if((stream->flags & _FILE_BUF_AUTO) && (stream->buf != NULL)) free(stream->buf);
	if(fscasio_struct(stream)->frags != NULL) free(fscasio_struct(stream)->frags);
	free(stream->private_data);

	free(stream);
	for (i=0; (i<FOPEN_MAX) && (__files[i] != stream); i++);
	if(__files[i] == stream) {
		__files[i] = NULL;
	}
	else return -1L;
}


int fscasio_remove(const char *filename) {

}


int fscasio_rename(const char *old, const char *new) {

}


int fscasio_fgetc(FILE *stream) {
	if((stream->flags & _FILE_EOF_REATCHED) || !(stream->open_mode & _FILE_READ)) return EOF;
	else {
		struct _fscasio_file *pdata = fscasio_struct(stream);
		if(pdata->total_size <= stream->pos) {
			stream->flags |= _FILE_EOF_REATCHED;
			return EOF;
		}
		else {
			int ret;
			ret = (int)(*(unsigned char*)((int)pdata->frags[pdata->current_frag].fstart + pdata->curfrag_pos));
			stream->pos++;
			pdata->curfrag_pos++;
			if(pdata->curfrag_pos >= pdata->frags[pdata->current_frag].fsize) {
				if(stream->pos >= pdata->total_size) {
					stream->flags |= _FILE_EOF_REATCHED;
					return EOF;
				}
				else {
					pdata->curfrag_pos = 0;
					pdata->current_frag++;
				}
			}
			return ret;
		}
	}
}


char *fscasio_fgets(char *s, int num, FILE *stream) {
	if(num==0 || !(stream->open_mode & _FILE_READ)) return s;
	else if((stream->flags & _FILE_EOF_REATCHED) || (stream->pos >= fscasio_struct(stream)->total_size)) {
		stream->flags |= _FILE_EOF_REATCHED;
		return NULL;
	}
	else {
		int maxRead;
		int nfrags;
		int i, j, n, k, oldSize, toRead;
		int posInFrag;
		int posInPtr;
		int fragSize;

		maxRead = fscasio_struct(stream)->total_size - stream->pos;
		maxRead = maxRead < num-1 ? maxRead : num-1;

		n = stream->pos + maxRead;
		j = stream->pos;
		i = fscasio_struct(stream)->current_frag;
		posInFrag = fscasio_struct(stream)->curfrag_pos;
		posInPtr = 0;
		while(j<n) {
			fragSize = fscasio_struct(stream)->frags[i].fsize - posInFrag;
			toRead = (j+fragSize) < n ? fragSize : n-j;
			for(k=0; k<toRead && (s[k+posInPtr] = *(char*)((int)(fscasio_struct(stream)->frags[i].fstart) + posInFrag + k)) != '\n'; k++);
			if(k!=toRead) k++;
			j+=k;
			posInPtr+=k;
			
			if(k == fragSize) {
				i++;
				posInFrag=0;
			}
			else posInFrag += k;

			if(s[posInPtr-1] == '\n') break;
		}

		stream->pos += posInPtr;
		fscasio_struct(stream)->current_frag = i;
		fscasio_struct(stream)->curfrag_pos = posInFrag;
		if(stream->pos >= fscasio_struct(stream)->total_size) stream->flags |= _FILE_EOF_REATCHED;
		s[posInPtr] = 0;
		return s;
	}
}


int fscasio_fputc(int c, FILE *stream) {

}


int fscasio_fputs(const char *s, FILE *stream) {

}


int fscasio_ungetc(int c, FILE *stream) {

}


size_t fscasio_fread(void *ptr, size_t size, size_t nmemb, FILE *stream) {
	if((size == 0) || (nmemb == 0) || (stream->flags & _FILE_EOF_REATCHED) || !(stream->open_mode & _FILE_READ)) {
		return 0;
	}
	else {
		int maxRead;
		int nfrags;
		int i, j, n, oldSize, toRead;
		int posInFrag;
		int posInPtr;
		int fragSize;

		maxRead = fscasio_struct(stream)->total_size - stream->pos;
		maxRead = maxRead<(size*nmemb) ? maxRead : (size*nmemb);

		n = stream->pos + maxRead;
		j = stream->pos;
		i = fscasio_struct(stream)->current_frag;
		posInFrag = fscasio_struct(stream)->curfrag_pos;
		posInPtr = 0;
		while(j<n) {
			fragSize = fscasio_struct(stream)->frags[i].fsize - posInFrag;
			toRead = (j+fragSize) < n ? fragSize : n-j;
			memcpy((void*)((int)ptr + posInPtr), (void*)((int)(fscasio_struct(stream)->frags[i].fstart) + posInFrag), toRead);
			j+=toRead;
			posInPtr+=toRead;
			
			if(toRead == fragSize) {
				i++;
				posInFrag=0;
			}
			else posInFrag += toRead;
		}

		stream->pos += maxRead;
		fscasio_struct(stream)->current_frag = i;
		fscasio_struct(stream)->curfrag_pos = posInFrag;
		if(stream->pos >= fscasio_struct(stream)->total_size) stream->flags |= _FILE_EOF_REATCHED;
		return maxRead;
	}
}


size_t fscasio_fwrite(const void *ptr, size_t size, size_t nmemb, FILE *stream) {

}


int fscasio_fseek(FILE *stream, long int offset, int whence) {
	int newpos;
	
	newpos = (whence == SEEK_CUR) ? stream->pos + offset : 
			( (whence == SEEK_SET) ? (int)offset : fscasio_struct(stream)->total_size + offset);

	if(newpos < 0 || newpos > fscasio_struct(stream)->total_size) return -1;
	else {
		int i, j, frag, posInFrag, oldPos;
		j=0;
		frag = -1;
		for(i=0; i<fscasio_struct(stream)->frag_num; i++) {
			oldPos = j;
			j += fscasio_struct(stream)->frags[i].fsize;
			if(j > newpos) {
				frag = i;
				posInFrag = newpos - oldPos;
				break;
			}
		}
		
		if(frag < 0) return -1;
		else {
			fscasio_struct(stream)->current_frag = frag;
			fscasio_struct(stream)->curfrag_pos = posInFrag;
			stream->pos = newpos;
			stream->flags &= ~(_FILE_EOF_REATCHED | _FILE_WRITTEN | _FILE_READED);
			return 0;
		}
	}
}


long int fscasio_ftell(FILE *stream) {
	return stream->pos;
}


int fscasio_fflush(FILE *stream) {

}







const unsigned char *getFSHeaderAddress() {
	unsigned char pageStartC;
	int i;

	pageStartC = CASIO_FS[0];
	for(i=0; CASIO_FS[i * HEADER_SIZE] == pageStartC; i++);
	return (const unsigned char *)((int)(CASIO_FS)+i*HEADER_SIZE);
}



const unsigned char *getAtomicFileHeader(const char *filename, unsigned short parentDirId) {
	const unsigned char *start, *current, *ret;
	int i;

	start = getFSHeaderAddress();
	current = start;

	for(i=0;(current[0] == FILE_HEADER_START) || (current[0] == FILE_HEADER_DELETED); i++) {
		if(!isDeleted(current)) {
			if(parentDirId == getFileParentDirId(current)) {
				if(cmpstr2fhn(filename, (const unsigned char*)((int)(current)+PREHEADER_SIZE), CHAR_CASE_AUTO) == 0) {
					return current;
				}
			}
		}
		if(!isDirectory(current)) current = 
			(const unsigned char*)((int)(current)+getFileFragNumber(current)*HEADER_SIZE + PREHEADER_SIZE + FILENAME_SIZE);
		else current = (const unsigned char*)((int)(current)+PREHEADER_SIZE + FILENAME_SIZE);
	}
	
	// Not found...
	return NULL;
}



const unsigned char *getFileHeader(const char *filename) {
	int i, j, start, n, stop;
	char s[30], sepchar;
	const unsigned char *ret, *tmp;
	unsigned short currentId;

	start=-1;
	ret = NULL;

	if(filename[0]=='\\' && filename[1]=='\\' && filename[2]=='f' && filename[3]=='l' && filename[4]=='s' && filename[5]=='0' && filename[6]=='\\') {
		start = 7;
		sepchar = '\\';
	}
	else if(filename[0]=='/' && filename[1]=='f' && filename[2]=='l' && filename[3]=='s' && filename[4]=='0' && filename[5]=='/') {
		start = 6;
		sepchar = '/';
	}

	if(start != -1) {
		n=start; 
		currentId = ROOT_ID;
		stop = 0;
		while(!stop) {
			for(i=0; (filename[i+n] != sepchar) && (filename[i+n] != 0) && (i<29); i++) s[i] = filename[i+n];
			s[i]=0;
			if(filename[i+n] == 0) {
				ret = getAtomicFileHeader(s, currentId);
				stop = 1;
			}
			else {
				tmp = getAtomicFileHeader(s, currentId);
				if(tmp == NULL) {
					ret = NULL;
					stop = 1;
				}
				else currentId = getFileId(tmp);
			}
			n += i+1;
		}
	}
	return ret;
}



int cmpstr2fhn(const char *str, const unsigned char *fileHeaderName, int caseRule) {
	int ret, i, j, k;
	char s[13];

	for(i=0; (str[i] != 0) && (str[i] != '.') && (i<8); i++) s[i]=str[i];
	if((i==8) && (str[8] != '.') && (str[8] != 0)) {
		s[6]='~';
		s[7]='1';
	}
	j=i;
	if((str[i] != '.') && (str[i] != 0)) for(j=i; (str[j] != '.') && (str[j] != 0); j++);
	if(str[j] != 0) for(k=0; (str[j+k] != 0) && (k<4); k++, i++) s[i]=str[j+k];
	s[i] = 0;
	
	j=1;
	ret = 0;
	for (i=0; i<12; i++) {
		if(s[i]==0) j=0;
		if(j) {
			if(caseRule == CHAR_CASE_AUTO) if((toupper(s[i]) != fileHeaderName[i*2+1]) && (tolower(s[i]) != fileHeaderName[i*2+1])) ret = -i-1;
			else if (caseRule == CHAR_CASE_UPPER) if(toupper(s[i]) != fileHeaderName[i*2+1]) ret = -i-1;
			// Else, don't change the character case (eq CHAR_CASE_NOCHANGE).		
			else if(s[i] != fileHeaderName[i*2+1]) ret = -i-1;
		}
		else if(fileHeaderName[i*2+1] != 0xFF) ret = -i-1;

//
		if(ret<0) return ret;
//
	}

	return ret;
}


int isDirectory(const unsigned char *fileHeader) {
	if(fileHeader[1] == FILE_HEADER_DIR) return 1;
	return 0;
}


int isDeleted(const unsigned char *fileHeader) {
	if(fileHeader[0] == FILE_HEADER_DELETED) return 1;
	return 0;
}


unsigned short getFileId(const unsigned char *fileHeader) {
	return *((unsigned short*)((int)(fileHeader)+2));
}


unsigned short getFileParentDirId(const unsigned char *fileHeader) {
	return *((unsigned short*)((int)(fileHeader)+6));
}


unsigned short getFileFragNumber(const unsigned char *fileHeader) {
	if(isDirectory(fileHeader)) return 0;
	else {
		return *((unsigned short*)((int)(fileHeader)+PREHEADER_SIZE+FILENAME_SIZE+10));
	}
}


unsigned int getFileTotalSize(const unsigned char *fileHeader) {
	unsigned short fnum = getFileFragNumber(fileHeader);
	if(fnum == 0) return 0;
	else {
		int i;
		unsigned int ret = 0;
		int pos = PREHEADER_SIZE+FILENAME_SIZE+18;
		for(i=0; i<fnum; i++, pos+=HEADER_SIZE) ret += *((unsigned short*)((int)(fileHeader)+pos))+1;
		return ret;
	}
}

//@TODO : This function never return -1 : add a 2nd condition in the 'for' loop.
unsigned int getPageOffsetByID(unsigned short id) {
	unsigned char pageStartC;
	int i, pos;

	pos = 0;
	pageStartC = CASIO_FS[0];
	for(i=0; CASIO_FS[pos] == pageStartC; i++, pos+=HEADER_SIZE) 
		if(*((unsigned short*)((int)(CASIO_FS)+pos+10)) == id) return *((unsigned int*)((int)(CASIO_FS)+pos+4));
	return -1;
}


const unsigned char *getFragDataAddress(const unsigned char *fragAdr) {
	unsigned int pageOff = getPageOffsetByID(*((unsigned short*)((int)(fragAdr)+14)));
	if(pageOff != -1) {
		pageOff += (int)(CASIO_STORAGE_MEM) + *((unsigned short*)((int)(fragAdr)+16));
		return (const unsigned char*)pageOff;
	}
	else return NULL;
}

